/*
 *  cMoHist.h
 *  MotionHistogram
 *
 *  Created by Kris Kitani on 5/25/11. (kkitani@cs.cmu.edu)
 *  Copyright 2011 Kris Kitani. All rights reserved.
 *
 */

#include "cMoHist.h"

using namespace std;

cMoHist::cMoHist(){
	
	numMagBins = 3;
	numVarBins = 3;
	numOriBins = 4;
	numFreqX = 8;
	numFreqY = 8;
	
	projection_error = 3; // 3
	
	avgmag_threshold = 1;
	var_threshold = 8;
	mag_threshold = 9;
	
	max_ori = 360;
	ori_bin_width = max_ori/numOriBins; 
	ori_shift = ori_bin_width/2;			
	dft_window_size = 64;
	
	t_freq_x = 200;
	t_freq_y = 200;
	
	norm_max_X = new float[numFreqX];
	
	norm_max_X[0]=1000;
	norm_max_X[1]=1600;
	norm_max_X[2]=700;
	norm_max_X[3]=600;
	norm_max_X[4]=600;
	norm_max_X[5]=600;
	norm_max_X[6]=600;
	norm_max_X[7]=600;
	
	norm_max_Y = new float[numFreqY];
	
	norm_max_Y[0]=800;
	norm_max_Y[1]=1000;
	norm_max_Y[2]=600;
	norm_max_Y[3]=600;
	norm_max_Y[4]=600;
	norm_max_Y[5]=600;
	norm_max_Y[6]=600;
	norm_max_Y[7]=600;
	
	
	for(int i=0;i< dft_window_size;i++){
		sigX.push_back(0);
		sigY.push_back(0);
		vsig.push_back(0);
	}
	
	src = cv::Mat(1,dft_window_size,CV_32F);
	_src = src;
	
	f=-64;
}

cMoHist::~cMoHist(){
	if(outstream.is_open()) outstream.close();
}

int cMoHist::getDims(){
	numDirectionalBins = numMagBins*numVarBins*numOriBins;
	numFreqBins = numFreqX + numFreqY;
	numTotalBins = numDirectionalBins + numFreqBins + 1; 
	return numTotalBins;
}

void cMoHist::computeMotionFeature(CvPoint2D32f *p, CvPoint2D32f *c, int count, double **motvec){
	
	if(count<4) return;
	
	
	// ==== FIND POINTS ON DOMINANT MOTION PLANE ==== //
	
	
	CvMat *orig, *dest, *homg;
	orig = cvCreateMat(2,count,CV_32FC1);
	dest = cvCreateMat(2,count,CV_32FC1);
	stat = cvCreateMat(1,count,CV_8UC1);
	homg = cvCreateMat(3,3,CV_32FC1);
	
	for(int i=0;i<count;i++){
		cvmSet(orig,0,i,p[i].x);
		cvmSet(orig,1,i,p[i].y);
		cvmSet(dest,0,i,c[i].x);
		cvmSet(dest,1,i,c[i].y);
	}
	
	cvFindHomography(orig, dest, homg, CV_RANSAC, projection_error, stat);
	
	///////////////////////////////////////
	//	     COMPUTE MOTION FEATURE      //
	///////////////////////////////////////
	
	float avg_mag = 0;
	float avg_mag_old = 0;
	float var = 0;
	int k=0, y=0, x=0;
	float avg_x = 0;
	float avg_y = 0;
	float avg_posX=0, avg_negX=0, avg_posY=0, avg_negY=0;
	int px=0,nx=0,py=0,ny=0;
	float dx,dy,mag;
	
	
	// ==== COMPUTE VARIANCE AND AVERAGE MAGNITUDE ==== //
	
	for(int i=0;i<count;i++){
		
		if(cvGetReal1D(stat,i) == 0) continue;
		
		dx = c[i].x - p[i].x;
		dy = c[i].y - p[i].y;
		mag = hypot(dy,dx);
		avg_mag_old = avg_mag; 
		avg_mag = (avg_mag*k + mag) / (k+1);
		var = ( k*var+(mag-avg_mag)*(mag-avg_mag_old) ) / (k+1);   // population variance
		k++;
		avg_x = (avg_x*x + dx) / (x+1);
		x++;
		avg_y = (avg_y*y + dy) / (y+1);
		y++;
		
		if(dx>0) { avg_posX = (avg_posX*px + dx) / (px+1); px++;}
		else{ avg_negX = (avg_negX*nx + dx) / (nx+1); nx++;}
		
		if(dy>0){ avg_posY = (avg_posY*py + dy) / (py+1); py++;}
		else{ avg_negY = (avg_negY*ny + dy) / (ny+1); ny++;}
	}
	
	// ==== GENERATE DIRECTIONAL VECTOR ==== //
	
	if(avg_mag<avgmag_threshold);
	else{
		float sum=0;
		for(int i=0;i<count;i++){
			if(cvGetReal1D(stat,i) == 0) continue;
			
			// ==== VARIANCE INDEX OFFSET ====
			int var_offset = -1;
			
			for(int m=numVarBins-1;m>=0;m--){
				if(var_offset!=-1) continue;
				else if(var>(var_threshold*m)) var_offset = cvRound(numOriBins*numMagBins*m);	
			}
			
			// ==== ORIENTATION INDEX OFFSET ====
			float dx = c[i].x - p[i].x;
			float dy = c[i].y - p[i].y;
			float ori = cvFastArctan(dy,dx);				// 0 - 360
			ori = ori + ori_shift;							// shift to avoid axes
			if(ori > max_ori) ori = ori - max_ori;			// account for over shifted values
			int   bin = cvFloor(ori/ori_bin_width);			// map to num_ori_bins
			
			// ==== MAGNITUDE INDEX OFFSET ====
			float mag = cvSqrt(dy*dy + dx*dx);
			
			int mag_offset = -1;
			for(int m=numMagBins-1;m>=0;m--){
				if(mag_offset>=0) continue;
				else if(mag>(mag_threshold*m)) mag_offset = cvRound(numOriBins*m);	
			}
			
			int index = bin + mag_offset + var_offset;
			motvec[63][index] += mag;
			sum += mag;
		}
		if(sum>0) for(int i=0;i<numDirectionalBins;i++) motvec[63][i] = motvec[63][i] * 1000/sum;
	}
	
	
	////////////////////////////////
	//     SPECTRAL ANALYSIS      //
	////////////////////////////////
	
	int i,j;
	
	sigX.push_back(avg_x);
	sigX.pop_front();
	copy(sigX.begin(), sigX.end(), vsig.begin());
	for(i=0;i<dft_window_size;i++) _src(0,i) = vsig[i];
	
	cv::dft(src,src,0,0);
	if(fabs(_src(0,0))>t_freq_x) motvec[31][numDirectionalBins] = fabs(1000*_src(0,0)/norm_max_X[0]);
	for(i=1,j=0; i<(numFreqX*2-1);i+=2,j++) if(fabs(_src(0,i))>t_freq_x) motvec[31][numDirectionalBins+j] = fabs(1000*_src(0,i)/norm_max_X[j]);
	
	// === Udate BOF (with Y Fourier spectra)
	sigY.push_back(avg_y);
	sigY.pop_front();
	copy(sigY.begin(), sigY.end(), vsig.begin());
	for(i=0;i<dft_window_size;i++) _src(0,i) = vsig[i];
	cv::dft(src,src,0,0);
	if(fabs(_src(0,0))>t_freq_y) motvec[31][numDirectionalBins+numFreqX] = fabs(1000*_src(0,0)/norm_max_Y[0]);
	for(i=1,j=0; i<(numFreqY*2-1);i+=2,j++) if(fabs(_src(0,i))>t_freq_y) motvec[31][numDirectionalBins+numFreqX+j] = fabs(1000*_src(0,i)/norm_max_Y[j]);
	
	double sum = 0;		
	for(int i=0;i<numTotalBins;i++){
		sum+= motvec[31][i];
	}
	if(sum==0) motvec[31][numTotalBins-1]=1000;
	
    //cvReleaseMat(&stat);
	cvReleaseMat(&orig);
	cvReleaseMat(&dest);
	cvReleaseMat(&homg);
}

void cMoHist::drawFlow(CvPoint2D32f *p, CvPoint2D32f *c, int count, IplImage *frm){
	
	for(int i=0;i<count;i++){
		if(cvGetReal1D(stat,i) == 0) continue;
		cvLine(frm,cvPoint(c[i].x,c[i].y),cvPoint(p[i].x,p[i].y),CV_RGB(255,255,0),1,CV_AA);
		cvCircle(frm,cvPoint(c[i].x,c[i].y),2,CV_RGB(0,255,0),-1,CV_AA);
	}
	cvShowImage("flow",frm);
}

void cMoHist::writeToFile(string filename, double **motvec){
	
	f++;
	if(outputfile!=filename){
		outputfile = filename;
		outstream.open(filename.c_str());
	}
	else if(f>0) {
		for(int i=0;i<numTotalBins;i++){
			outstream << cvRound(motvec[31][i]) << " ";
		}
		outstream << endl;
	}
	else{} // skip first 64 frames
	
	// ==== SHIFT MEMORY ==== //
	for(int i=1;i<64;i++)
		for(int j=0;j<numTotalBins;j++)
			motvec[i-1][j] = motvec[i][j];
	
	memset(motvec[63], 0, numTotalBins*sizeof(double));
}
